use crossbeam::channel::Receiver;
use nalgebra::vector;
use rapier2d::{
	dynamics::{RigidBody, RigidBodyBuilder, RigidBodyHandle},
	geometry::{ColliderBuilder, ColliderHandle, CollisionEvent},
	pipeline::ChannelEventCollector,
};
use std::collections::HashMap;

use super::{
	physicswrapper::PhysicsWrapper, Projectile, ProjectileBuilder, Ship, ShipHandle, ShipOutfit,
};
use crate::{content, render::Sprite};

/// Keeps track of all objects in the world that we can interact with.
/// Also wraps our physics engine
pub struct World {
	physics: PhysicsWrapper,
	projectiles: HashMap<ColliderHandle, Projectile>,
	ships: HashMap<ColliderHandle, Ship>,

	collision_handler: ChannelEventCollector,
	collision_queue: Receiver<CollisionEvent>,
}

// Private methods
impl World {
	fn remove_projectile(&mut self, c: ColliderHandle) {
		let p = match self.projectiles.remove(&c) {
			Some(p) => p,
			None => return,
		};
		self.physics.rigid_body_set.remove(
			p.rigid_body,
			&mut self.physics.im,
			&mut self.physics.collider_set,
			&mut self.physics.ij,
			&mut self.physics.mj,
			true,
		);
	}

	fn remove_ship(&mut self, h: ShipHandle) {
		self.physics.rigid_body_set.remove(
			h.0,
			&mut self.physics.im,
			&mut self.physics.collider_set,
			&mut self.physics.ij,
			&mut self.physics.mj,
			true,
		);
		self.ships.remove(&h.1);
	}

	fn add_projectile(&mut self, pb: ProjectileBuilder) -> ColliderHandle {
		let r = self.physics.rigid_body_set.insert(pb.rigid_body.build());
		let c = self.physics.collider_set.insert_with_parent(
			pb.collider.build(),
			r,
			&mut self.physics.rigid_body_set,
		);
		self.projectiles.insert(c, pb.build(r, c));
		return c;
	}
}

// Public methods
impl World {
	pub fn new() -> Self {
		let (collision_send, collision_queue) = crossbeam::channel::unbounded();
		let (contact_force_send, _) = crossbeam::channel::unbounded();

		Self {
			physics: PhysicsWrapper::new(),
			projectiles: HashMap::new(),
			ships: HashMap::new(),
			collision_handler: ChannelEventCollector::new(collision_send, contact_force_send),
			collision_queue,
		}
	}

	pub fn add_ship(&mut self, ct: &content::Ship, outfits: Vec<ShipOutfit>) -> ShipHandle {
		let rb = RigidBodyBuilder::dynamic()
			.translation(vector![0.0, 0.0])
			.can_sleep(false);
		let cl = ColliderBuilder::ball(50.0).restitution(0.7).mass(1.0);

		let r = self.physics.rigid_body_set.insert(rb.build());
		let c = self.physics.collider_set.insert_with_parent(
			cl.build(),
			r,
			&mut self.physics.rigid_body_set,
		);

		let h = ShipHandle(r, c);
		self.ships.insert(c, Ship::new(ct, outfits, h));
		return h;
	}

	pub fn tick(&mut self, t: f32) {
		// Run ship updates
		let mut res = Vec::new();
		let mut to_remove = Vec::new();
		for (_, s) in &mut self.ships {
			if s.hull <= 0.0 {
				to_remove.push(s.physics_handle);
				continue;
			}
			let r = &mut self.physics.rigid_body_set[s.physics_handle.0];
			res.push(s.tick(r, t));
		}
		for r in to_remove {
			self.remove_ship(r);
		}
		for r in res {
			for p in r.projectiles {
				self.add_projectile(p);
			}
		}

		// Update physics
		self.physics.step(t, &self.collision_handler);

		// Handle collision events
		while let Ok(event) = &self.collision_queue.try_recv() {
			if event.started() {
				let a = &event.collider1();
				let b = &event.collider2();

				// If a projectile is part of the collision, make sure
				// a is a projectile.
				let (a, b) = if self.projectiles.contains_key(b) {
					(b, a)
				} else {
					(a, b)
				};

				if let Some(p) = self.projectiles.get(a) {
					if let Some(s) = self.ships.get_mut(b) {
						s.hull -= p.damage;
					}
					self.remove_projectile(*a);
					self.remove_projectile(*b);
				}
			}
		}

		// Delete projectiles
		let mut to_remove = Vec::new();
		for (_, p) in &mut self.projectiles {
			p.tick(t);
			if p.is_expired() {
				to_remove.push(p.collider);
			}
		}
		for i in to_remove {
			self.remove_projectile(i);
		}
	}

	pub fn get_rigid_body(&self, r: RigidBodyHandle) -> &RigidBody {
		&self.physics.rigid_body_set[r]
	}

	pub fn get_ship_mut(&mut self, s: &ShipHandle) -> &mut Ship {
		self.ships.get_mut(&s.1).unwrap()
	}

	pub fn get_ship_sprites(&self) -> impl Iterator<Item = Sprite> + '_ {
		self.ships
			.values()
			.map(|x| x.get_sprite(&self.physics.rigid_body_set[x.physics_handle.0]))
	}

	pub fn get_projectile_sprites(&self) -> impl Iterator<Item = Sprite> + '_ {
		self.projectiles
			.values()
			.map(|x| x.get_sprite(&self.physics.rigid_body_set[x.rigid_body]))
	}
}
